@@ -1,6 +1,6 @@
 #!/usr/bin/python
 #
-# Copyright (C) 2006 Google Inc.
+# Copyright (C) 2006, 2007, 2008 Google Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -21,6 +21,10 @@
                operations with the Atom Publishing Protocol on which GData is
                based. An instance can perform query, insertion, deletion, and
                update.
+
+  HttpRequest: Function that performs a GET, POST, PUT, or DELETE HTTP request
+       to the specified end point. An AtomService object or a subclass can be
+       used to specify information about the request.
 """
 
 __author__ = 'api.jscudder (Jeffrey Scudder)'
@@ -37,10 +41,15 @@ except ImportError:
   try:
     import cElementTree as ElementTree
   except ImportError:
-    from elementtree import ElementTree
+    try:
+      from xml.etree import ElementTree
+    except ImportError:
+      from elementtree import ElementTree
+
 
 URL_REGEX = re.compile('http(s)?\://([\w\.-]*)(\:(\d+))?(/.*)?')
 
+
 class AtomService(object):
   """Performs Atom Publishing Protocol CRUD operations.
   
@@ -72,34 +81,7 @@ class AtomService(object):
   def _ProcessUrl(self, url, for_proxy=False):
     """Processes a passed URL.  If the URL does not begin with https?, then
     the default value for self.server is used"""
-
-    server = self.server
-    if for_proxy:
-      port = 80
-      ssl = False
-    else:
-      port = self.port
-      ssl = self.ssl
-    uri = url
-
-    m = URL_REGEX.match(url)
-
-    if m is None:
-      return (server, port, ssl, uri)
-    else:
-      if m.group(1) is not None:
-        port = 443
-        ssl = True
-      if m.group(3) is None:
-        server = m.group(2)
-      else:
-        server = m.group(2)
-        port = int(m.group(4))
-      if m.group(5) is not None:
-        uri = m.group(5)
-      else:
-        uri = '/'
-      return (server, port, ssl, uri)
+    return ProcessUrl(self, url, for_proxy=for_proxy)
 
   def UseBasicAuth(self, username, password, for_proxy=False):
     """Sets an Authenticaiton: Basic HTTP header containing plaintext.
@@ -112,14 +94,7 @@ class AtomService(object):
       username: str
       password: str
     """
-
-    base_64_string = base64.encodestring('%s:%s' % (username, password))
-    base_64_string = base_64_string.strip()
-    if for_proxy:
-      header_name = 'Proxy-Authorization'
-    else:
-      header_name = 'Authorization'
-    self.additional_headers[header_name] = 'Basic %s' % (base_64_string,)
+    UseBasicAuth(self, username, password, for_proxy=for_proxy)
 
   def PrepareConnection(self, full_uri):
     """Opens a connection to the server based on the full URI.
@@ -138,114 +113,12 @@ class AtomService(object):
       A tuple containing the httplib.HTTPConnection and the full_uri for the
       request.
     """
-    
-    (server, port, ssl, partial_uri) = self._ProcessUrl(full_uri)
-    if ssl:
-      # destination is https
-      proxy = os.environ.get('https_proxy')
-      if proxy:
-        (p_server, p_port, p_ssl, p_uri) = self._ProcessUrl(proxy, True)
-        proxy_username = os.environ.get('proxy-username')
-        if not proxy_username:
-          proxy_username = os.environ.get('proxy_username')
-        proxy_password = os.environ.get('proxy-password')
-        if not proxy_password:
-          proxy_password = os.environ.get('proxy_password')
-        if proxy_username:
-          user_auth = base64.encodestring('%s:%s' % (proxy_username, 
-                                                     proxy_password))
-          proxy_authorization = ('Proxy-authorization: Basic %s\r\n' % (
-              user_auth.strip()))
-        else:
-          proxy_authorization = ''
-        proxy_connect = 'CONNECT %s:%s HTTP/1.0\r\n' % (server,port)
-        user_agent = 'User-Agent: %s\r\n' % (
-            self.additional_headers['User-Agent'])
-        proxy_pieces = (proxy_connect + proxy_authorization + user_agent 
-                        + '\r\n')
-
-        #now connect, very simple recv and error checking
-        p_sock = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
-        p_sock.connect((p_server,p_port))
-        p_sock.sendall(proxy_pieces)
-        response = ''
-
-        # Wait for the full response.
-        while response.find("\r\n\r\n") == -1:
-          response += p_sock.recv(8192)
-        
-        p_status=response.split()[1]
-        if p_status!=str(200):
-          raise 'Error status=',str(p_status)
-
-        # Trivial setup for ssl socket.
-        ssl = socket.ssl(p_sock, None, None)
-        fake_sock = httplib.FakeSocket(p_sock, ssl)
-
-        # Initalize httplib and replace with the proxy socket.
-        connection = httplib.HTTPConnection(server)
-        connection.sock=fake_sock
-        full_uri = partial_uri
-
-      else:
-        connection = httplib.HTTPSConnection(server, port)
-        full_uri = partial_uri
-
-    else:
-      # destination is http
-      proxy = os.environ.get('http_proxy')
-      if proxy:
-        (p_server, p_port, p_ssl, p_uri) = self._ProcessUrl(proxy, True)
-        proxy_username = os.environ.get('proxy-username')
-        if not proxy_username:
-          proxy_username = os.environ.get('proxy_username')
-        proxy_password = os.environ.get('proxy-password')
-        if not proxy_password:
-          proxy_password = os.environ.get('proxy_password')
-        if proxy_username:
-          self.UseBasicAuth(proxy_username, proxy_password, True)
-        connection = httplib.HTTPConnection(p_server, p_port)
-        if not full_uri.startswith("http://"):
-          if full_uri.startswith("/"):
-            full_uri = "http://%s%s" % (self.server, full_uri)
-          else:
-            full_uri = "http://%s/%s" % (self.server, full_uri)
-      else:
-        connection = httplib.HTTPConnection(server, port)
-        full_uri = partial_uri
-
-    return (connection, full_uri)
+    return PrepareConnection(self, full_uri)   
 
   # Alias the old name for the above method to preserve backwards 
   # compatibility.
   _PrepareConnection = PrepareConnection
  
-  def CreateConnection(self, uri, http_operation, extra_headers=None,
-      url_params=None, escape_params=True):
-      
-    full_uri = BuildUri(uri, url_params, escape_params)
-    (connection, full_uri) = self._PrepareConnection(full_uri)
-    connection.putrequest(http_operation, full_uri)
-
-    if isinstance(self.additional_headers, dict):
-      for header in self.additional_headers:
-        connection.putheader(header, self.additional_headers[header])
-    if isinstance(extra_headers, dict):
-      for header in extra_headers:
-        connection.putheader(header, extra_headers[header])
-    connection.endheaders()
-
-    # Turn on debug mode if the debug member is set
-    if self.debug:
-      connection.debuglevel = 1
-
-    return connection
-
-  # Set an alias to allow calling of _CreateConnection for backwards
-  # compatibility
-  _CreateConnection = CreateConnection
-
-
   # CRUD operations
   def Get(self, uri, extra_headers=None, url_params=None, escape_params=True):
     """Query the APP server with the given URI
@@ -280,11 +153,8 @@ class AtomService(object):
     Returns:
       httplib.HTTPResponse The server's response to the GET request.
     """
-
-    query_connection = self._CreateConnection(uri, 'GET', extra_headers,
-        url_params, escape_params)
-
-    return query_connection.getresponse()
+    return HttpRequest(self, 'GET', None, uri, extra_headers=extra_headers, 
+        url_params=url_params, escape_params=escape_params)
 
   def Post(self, data, uri, extra_headers=None, url_params=None, 
            escape_params=True, content_type='application/atom+xml'):
@@ -311,19 +181,9 @@ class AtomService(object):
     Returns:
       httplib.HTTPResponse Server's response to the POST request.
     """
-    if ElementTree.iselement(data):
-      data_str = ElementTree.tostring(data)
-    else:
-      data_str = str(data)
-    
-    extra_headers['Content-Length'] = len(data_str)
-    extra_headers['Content-Type'] = content_type
-    insert_connection = self._CreateConnection(uri, 'POST', extra_headers,
-        url_params, escape_params)
-
-    insert_connection.send(data_str)
-
-    return insert_connection.getresponse()
+    return HttpRequest(self, 'POST', data, uri, extra_headers=extra_headers, 
+        url_params=url_params, escape_params=escape_params, 
+        content_type=content_type)
 
   def Put(self, data, uri, extra_headers=None, url_params=None, 
            escape_params=True, content_type='application/atom+xml'):
@@ -350,19 +210,9 @@ class AtomService(object):
     Returns:
       httplib.HTTPResponse Server's response to the PUT request.
     """
-    if ElementTree.iselement(data):
-      data_str = ElementTree.tostring(data)
-    else:
-      data_str = str(data)
-      
-    extra_headers['Content-Length'] = len(data_str)
-    extra_headers['Content-Type'] = content_type
-    update_connection = self._CreateConnection(uri, 'PUT', extra_headers,
-        url_params, escape_params)
-
-    update_connection.send(data_str)
-
-    return update_connection.getresponse()
+    return HttpRequest(self, 'PUT', data, uri, extra_headers=extra_headers, 
+        url_params=url_params, escape_params=escape_params, 
+        content_type=content_type)
 
   def Delete(self, uri, extra_headers=None, url_params=None, 
              escape_params=True):
@@ -387,10 +237,296 @@ class AtomService(object):
     Returns:
       httplib.HTTPResponse Server's response to the DELETE request.
     """
-    delete_connection = self._CreateConnection(uri, 'DELETE', extra_headers,
-        url_params, escape_params)
+    return HttpRequest(self, 'DELETE', None, uri, extra_headers=extra_headers, 
+        url_params=url_params, escape_params=escape_params)
+
+
+def HttpRequest(service, operation, data, uri, extra_headers=None, 
+    url_params=None, escape_params=True, content_type='application/atom+xml'):
+  """Performs an HTTP call to the server, supports GET, POST, PUT, and DELETE.
+
+  Usage example, perform and HTTP GET on http://www.google.com/:
+    import atom.service
+    client = atom.service.AtomService()
+    http_response = client.Get('http://www.google.com/')
+  or you could set the client.server to 'www.google.com' and use the 
+  following:
+    client.server = 'www.google.com'
+    http_response = client.Get('/')
+
+  Args:
+    service: atom.AtomService object which contains some of the parameters 
+        needed to make the request. The following members are used to 
+        construct the HTTP call: server (str), additional_headers (dict), 
+        port (int), and ssl (bool).
+    operation: str The HTTP operation to be performed. This is usually one of
+        'GET', 'POST', 'PUT', or 'DELETE'
+    data: ElementTree, filestream, list of parts, or other object which can be 
+        converted to a string. 
+        Should be set to None when performing a GET or PUT.
+        If data is a file-like object which can be read, this method will read
+        a chunk of 100K bytes at a time and send them. 
+        If the data is a list of parts to be sent, each part will be evaluated
+        and sent.
+    uri: The beginning of the URL to which the request should be sent. 
+        Examples: '/', '/base/feeds/snippets', 
+        '/m8/feeds/contacts/default/base'
+    extra_headers: dict of strings. HTTP headers which should be sent
+        in the request. These headers are in addition to those stored in 
+        service.additional_headers.
+    url_params: dict of strings. Key value pairs to be added to the URL as
+        URL parameters. For example {'foo':'bar', 'test':'param'} will 
+        become ?foo=bar&test=param.
+    escape_params: bool default True. If true, the keys and values in 
+        url_params will be URL escaped when the form is constructed 
+        (Special characters converted to %XX form.)
+    content_type: str The MIME type for the data being sent. Defaults to
+        'application/atom+xml', this is only used if data is set.
+  """
+  full_uri = BuildUri(uri, url_params, escape_params)
+  (connection, full_uri) = PrepareConnection(service, full_uri)
+
+  if extra_headers is None:
+    extra_headers = {}
+
+  # Turn on debug mode if the debug member is set.
+  if service.debug:
+    connection.debuglevel = 1
+
+  connection.putrequest(operation, full_uri)
+
+  # If the list of headers does not include a Content-Length, attempt to 
+  # calculate it based on the data object.
+  if (data and not service.additional_headers.has_key('Content-Length') and 
+      not extra_headers.has_key('Content-Length')):
+    content_length = __CalculateDataLength(data)
+    if content_length:
+      extra_headers['Content-Length'] = content_length
+
+  if content_type:
+    extra_headers['Content-Type'] = content_type 
+
+  # Send the HTTP headers.
+  if isinstance(service.additional_headers, dict):
+    for header in service.additional_headers:
+      connection.putheader(header, service.additional_headers[header])
+  if isinstance(extra_headers, dict):
+    for header in extra_headers:
+      connection.putheader(header, extra_headers[header])
+  connection.endheaders()
+
+  # If there is data, send it in the request.
+  if data:
+    if isinstance(data, list):
+      for data_part in data:
+        __SendDataPart(data_part, connection)
+    else:
+      __SendDataPart(data, connection)
+
+  # Return the HTTP Response from the server.
+  return connection.getresponse()
+
+
+def __SendDataPart(data, connection):
+  if isinstance(data, str):
+    #TODO add handling for unicode.
+    connection.send(data)
+    return
+  elif ElementTree.iselement(data):
+    connection.send(ElementTree.tostring(data))
+    return
+  # Check to see if data is a file-like object that has a read method.
+  elif hasattr(data, 'read'):
+    # Read the file and send it a chunk at a time.
+    while 1:
+      binarydata = data.read(100000)
+      if binarydata == '': break
+      connection.send(binarydata)
+    return
+  else:
+    # The data object was not a file.
+    # Try to convert to a string and send the data.
+    connection.send(str(data))
+    return
+
+
+def __CalculateDataLength(data):
+  """Attempts to determine the length of the data to send. 
+  
+  This method will respond with a length only if the data is a string or
+  and ElementTree element.
+
+  Args:
+    data: object If this is not a string or ElementTree element this funtion
+        will return None.
+  """
+  if isinstance(data, str):
+    return len(data)
+  elif isinstance(data, list):
+    return None
+  elif ElementTree.iselement(data):
+    return len(ElementTree.tostring(data))
+  elif hasattr(data, 'read'):
+    # If this is a file-like object, don't try to guess the length.
+    return None
+  else:
+    return len(str(data))
+
+
+def PrepareConnection(service, full_uri):
+  """Opens a connection to the server based on the full URI.
+
+  Examines the target URI and the proxy settings, which are set as
+  environment variables, to open a connection with the server. This
+  connection is used to make an HTTP request.
+
+  Args:
+    service: atom.AtomService or a subclass. It must have a server string which
+      represents the server host to which the request should be made. It may also
+      have a dictionary of additional_headers to send in the HTTP request.
+    full_uri: str Which is the target relative (lacks protocol and host) or
+    absolute URL to be opened. Example:
+    'https://www.google.com/accounts/ClientLogin' or
+    'base/feeds/snippets' where the server is set to www.google.com.
+
+  Returns:
+    A tuple containing the httplib.HTTPConnection and the full_uri for the
+    request.
+  """
+   
+  (server, port, ssl, partial_uri) = ProcessUrl(service, full_uri)
+  if ssl:
+    # destination is https
+    proxy = os.environ.get('https_proxy')
+    if proxy:
+      (p_server, p_port, p_ssl, p_uri) = ProcessUrl(service, proxy, True)
+      proxy_username = os.environ.get('proxy-username')
+      if not proxy_username:
+        proxy_username = os.environ.get('proxy_username')
+      proxy_password = os.environ.get('proxy-password')
+      if not proxy_password:
+        proxy_password = os.environ.get('proxy_password')
+      if proxy_username:
+        user_auth = base64.encodestring('%s:%s' % (proxy_username,
+                                                   proxy_password))
+        proxy_authorization = ('Proxy-authorization: Basic %s\r\n' % (
+            user_auth.strip()))
+      else:
+        proxy_authorization = ''
+      proxy_connect = 'CONNECT %s:%s HTTP/1.0\r\n' % (server, port)
+      user_agent = 'User-Agent: %s\r\n' % (
+          service.additional_headers['User-Agent'])
+      proxy_pieces = (proxy_connect + proxy_authorization + user_agent
+                       + '\r\n')
+
+      #now connect, very simple recv and error checking
+      p_sock = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
+      p_sock.connect((p_server,p_port))
+      p_sock.sendall(proxy_pieces)
+      response = ''
+
+      # Wait for the full response.
+      while response.find("\r\n\r\n") == -1:
+        response += p_sock.recv(8192)
+       
+      p_status=response.split()[1]
+      if p_status!=str(200):
+        raise 'Error status=',str(p_status)
+
+      # Trivial setup for ssl socket.
+      ssl = socket.ssl(p_sock, None, None)
+      fake_sock = httplib.FakeSocket(p_sock, ssl)
+
+      # Initalize httplib and replace with the proxy socket.
+      connection = httplib.HTTPConnection(server)
+      connection.sock=fake_sock
+      full_uri = partial_uri
+
+    else:
+      connection = httplib.HTTPSConnection(server, port)
+      full_uri = partial_uri
+
+  else:
+    # destination is http
+    proxy = os.environ.get('http_proxy')
+    if proxy:
+      (p_server, p_port, p_ssl, p_uri) = ProcessUrl(service.server, proxy, True)
+      proxy_username = os.environ.get('proxy-username')
+      if not proxy_username:
+        proxy_username = os.environ.get('proxy_username')
+      proxy_password = os.environ.get('proxy-password')
+      if not proxy_password:
+        proxy_password = os.environ.get('proxy_password')
+      if proxy_username:
+        UseBasicAuth(service, proxy_username, proxy_password, True)
+      connection = httplib.HTTPConnection(p_server, p_port)
+      if not full_uri.startswith("http://"):
+        if full_uri.startswith("/"):
+          full_uri = "http://%s%s" % (service.server, full_uri)
+        else:
+          full_uri = "http://%s/%s" % (service.server, full_uri)
+    else:
+      connection = httplib.HTTPConnection(server, port)
+      full_uri = partial_uri
+
+  return (connection, full_uri)
+
+
+def UseBasicAuth(service, username, password, for_proxy=False):
+  """Sets an Authenticaiton: Basic HTTP header containing plaintext.
+  
+  The username and password are base64 encoded and added to an HTTP header
+  which will be included in each request. Note that your username and 
+  password are sent in plaintext. The auth header is added to the 
+  additional_headers dictionary in the service object.
+
+  Args:
+    service: atom.AtomService or a subclass which has an 
+        additional_headers dict as a member.
+    username: str
+    password: str
+  """
+  base_64_string = base64.encodestring('%s:%s' % (username, password))
+  base_64_string = base_64_string.strip()
+  if for_proxy:
+    header_name = 'Proxy-Authorization'
+  else:
+    header_name = 'Authorization'
+  service.additional_headers[header_name] = 'Basic %s' % (base_64_string,)
+
+
+def ProcessUrl(service, url, for_proxy=False):
+  """Processes a passed URL.  If the URL does not begin with https?, then
+  the default value for server is used"""
+
+  server = service.server
+  if for_proxy:
+    port = 80
+    ssl = False
+  else:
+    port = service.port
+    ssl = service.ssl
+  uri = url
+
+  m = URL_REGEX.match(url)
+
+  if m is None:
+    return (server, port, ssl, uri)
+  else:
+    if m.group(1) is not None:
+      port = 443
+      ssl = True
+    if m.group(3) is None:
+      server = m.group(2)
+    else:
+      server = m.group(2)
+      port = int(m.group(4))
+    if m.group(5) is not None:
+      uri = m.group(5)
+    else:
+      uri = '/'
+    return (server, port, ssl, uri)
 
-    return delete_connection.getresponse()
 
 def DictionaryToParamList(url_parameters, escape_params=True):
   """Convert a dictionary of URL arguments into a URL parameter string.
